/*
 * Implementation of "Visualization of Large Terrains Made Easy"
 * (Lindstrom and Pascucci) with geomorphing, embedded quadtree
 * indexing, view frustum culling, and multithreaded refinement
 *
 * Copyright (c) 2002 Jan Wassenberg
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Contact info:
 *   Jan.Wassenberg@stud.uni-karlsruhe.de
 *   http://www.stud.uni-karlsruhe.de/~urkt/
 */

#include "precompiled.h"

#define WIN32_PERF_MON

#include <SDL/sdl_thread.h>
#include "wsdl.h"

#include "system.h"
#include "res.h"
#include "util.h"		/* log2(), fast ceil() */


/* disable C++ name mangling (easier interface to external asm code) */
extern "C" {


#define _nu	97.78f	/* = xres / fov / tolerance */

/* dmax = e*nu[0]+r, dmin = e*nu[1]+r */
float nu[2] = { _nu, _nu * 2.f/3.f };

/*
 * is VAR available?
 * set by lod_init() if GL_NV_vertex_array_range is advertised
 */
static uint use_VAR;

/* preprocessed terrain vertex */
typedef struct
{
	int16 r;	/* radius enclosing all descendants */
	int8 dz;	/* z offset from z midpoint of triangle base points */
	uint8 e;	/* error */
}
V;
V *v;

/* terrain vertex coords (packed into 8 bytes for easier copying) */
typedef struct
{
	int16 x, y;
	float z;
}
VP;
VP vp[9];	/* coords of nodes 0-9 */


/*****************************************************************************
 * preprocessing
 *****************************************************************************/

static int init();

static uint w;	/* w == 2^(depth/2) + 1 */
uint depth;
static uint v_entries;
static const uchar* h;

/* stripped down version of V for preprocessing */
typedef struct
{
	int16 r;
	uint8 e;
}
PP_V;
static PP_V* pp_v;

/*
 * calculate radius of a sphere about m enclosing c and c's children
 * radius = (distance(m, c) + c's radius)
 */
static int xy_dist_sq = 1;	/* parent's xy distance ^2 to its children */

static inline int radius(const uint m, const uint c)
{
	const int dz = h[m]-h[c];
	const int dist = ceil(sqrtf((float)(xy_dist_sq+dz*dz)));

	return dist+pp_v[c].r;
}


#define MAX_E	(uint8)max(abs(h[m] - (h[b[0]]+h[b[1]])/2),\
				  max(max(pp_v[c1].e, pp_v[c2].e), max(pp_v[c3].e, pp_v[c4].e)))
#define MAX_R	(int16)max(max(radius(m,c1), radius(m,c2)),\
				  max(radius(m,c3), radius(m,c4)))


/*
 * calculate z-error and radius for each point on the terrain
 *
 * algorithm:
 *   for each level, bottom up
 *     for each point in the current level:
 *       error = max(current_error, max error of a child node)
 *       radius = max d (d = distance to a child node + child's radius)
 *
 * process whole levels (= all points at a certain depth in the qtree)
 * at once, bottom up; this means each point is only processed once.
 *
 * this requires access to all 4 (quadtree) children,
 * so we need the data in matrix order.
 * points in a level are then separated by xstep/ystep units;
 * xstep = ystep = 2 for the first even level.
 * xstep / ystep double before each odd / even level.
 *
 * points on an even level lie on a regular grid; each point has 4 children.
 * odd levels have every 2nd row shifted ystep to the right;
 * edge / non-edge points have 2 / 4 children, respectively.
 * the level loop is unrolled once to handle even/odd levels separately;
 * the "child valid" tests in odd levels are eliminated by
 * special casing the edge points.
 *
 * each loop carries along the children and base points
 * (used to calculate error) for speed.
 */
static void calc_e_r()
{
	/* per level: */
	uint start = 1;			/* x/y coord of first point in a level */
	uint odd_start = 2*w+4;	/* current odd level's first non-edge point */
	uint xstep = 2, ystep;	/* x/y distance to next point in level */

	/* counters */
	uint x, y;
	uint count = (w-1)/2;	/* # points per row / rows per level (even) */

	/* current point */
	uint m;					/* current point */
	uint b[2];				/* base points of m's triangle */
	uint c1, c2, c3, c4;	/* m's children */

	/*
	 * for each level:
	 * note: loop body handles 1 even and 1 odd level;
	 *   there are depth-1 levels, and the bottom (odd) level
	 *   is skipped (its d and r values are 0), so loop (depth-2)/2 times
	 */
	for(uint i = (depth-2)/2; i > 0; i--)
	{
		ystep = xstep;

		/*********************************************************************
		 * even level
		 * points are a regular grid; children are always valid
		 *********************************************************************/

		m = (w+1)*start;
		b[0] = 0; b[1] = 2*m;
		c1 = start;						/*   N   */
		c2 = start*w; c3 = start*(w+2);	/* W   E */
		c4 = start*(2*w+1);				/*   S   */

		/* for each row: */
		for(y = count; y > 0; y--)
		{
			/* for each point in row: */
			for(x = count; x > 0; x--)
			{
				pp_v[m].e = MAX_E;
				pp_v[m].r = MAX_R;

				/* go to next point in row */
				m += xstep;
				b[x&1] += 2*xstep;	/* \/\/ "spider walk" */
				c1 += xstep; c2 += xstep;	/* alternative: new c2 = old c3 */
				c3 += xstep; c4 += xstep;
			}

			/* go to next row */
			uint next_row_ofs = (xstep-1)*w+1;
			m += next_row_ofs;
			b[0] += next_row_ofs; b[1] += next_row_ofs;
			b[y&1] += xstep*w; b[1-(y&1)] -= xstep*w;
			c1 += next_row_ofs; c2 += next_row_ofs;	c3 += next_row_ofs;	c4 += next_row_ofs;
		}

		xstep += xstep;
		start += start;
		count >>= 1;
		xy_dist_sq += xy_dist_sq;	/* children are diagonally distanced */

		/* offset to the left/right child */
		const uint cl = (+w-1)*xstep/4, cr = (+w+1)*xstep/4;

		/*********************************************************************
		 * odd edge
		 * points are regularly spaced but have only 2 children
		 * special cased to simplify the main odd loop below
		 *********************************************************************/
		{
			/*             N          S           W           E         */
			uint m[4] = {start, start+(w-1)*w, start*w, start*w+(w-1)};

			for(x = count; x > 0; x--) /* do all 4 edges at once */
			{
				/* calculate error */
				pp_v[m[0]].e = (uchar)max(abs(h[m[0]] - (h[m[0]-ystep]+h[m[0]+ystep])/2), max(pp_v[m[0]+cl].e, pp_v[m[0]+cr].e));
				pp_v[m[1]].e = (uchar)max(abs(h[m[1]] - (h[m[1]-ystep]+h[m[1]+ystep])/2), max(pp_v[m[1]-cr].e, pp_v[m[1]-cl].e));
				pp_v[m[2]].e = (uchar)max(abs(h[m[2]] - (h[m[2]-ystep*w]+h[m[2]+ystep*w])/2), max(pp_v[m[2]-cl].e, pp_v[m[2]+cr].e));
				pp_v[m[3]].e = (uchar)max(abs(h[m[3]] - (h[m[3]-ystep*w]+h[m[3]+ystep*w])/2), max(pp_v[m[3]-cr].e, pp_v[m[3]+cl].e));

				/*
				 * calculate radius such that all descendants
				 * are contained in the resulting sphere
				 */
				pp_v[m[0]].r = (short)max(radius(m[0],m[0]+cl), radius(m[0],m[0]+cr));
				pp_v[m[1]].r = (short)max(radius(m[1],m[1]-cr), radius(m[1],m[1]-cl));
				pp_v[m[2]].r = (short)max(radius(m[2],m[2]-cl), radius(m[2],m[2]+cr));
				pp_v[m[3]].r = (short)max(radius(m[3],m[3]-cr), radius(m[3],m[3]+cl));

				/* go to next point on all edges */
				m[0] += xstep; m[1] += xstep; m[2] += xstep*w; m[3] += xstep*w;
			}
		}

		/*********************************************************************
		 * odd level (excluding edges)
		 * every 2nd row is shifted ystep to the right; all children are valid
		 *********************************************************************/

		m = odd_start;
		b[0] = m-ystep*w; b[1] = m+ystep*w;
		c1 = m-cr; c2 = c1+ystep;	/* NW NE */
		c3 = m+cl; c4 = c3+ystep;	/* SW SE */

		/* for each row: */
		for(y = 2*count-1; y > 0; y--)	/* N, S edges already done */
		{
			/*
			 * for each point in row:
			 * NOTE: all points on the highest level are on an edge;
			 *       this loop is skipped by the x = ...; x > 0 statement
			 */
			for(x = count-(y&1); x > 0; x--) /* 1 point less every other row */
			{
				pp_v[m].e = MAX_E;
				pp_v[m].r = MAX_R;

				/* go to next point in row */
				m += xstep;
				b[0] += xstep; b[1] += xstep;
				c1 += xstep; c2 += xstep; c3 += xstep; c4 += xstep;
			}

			/* go to next row */
			uint next_row_ofs = (ystep-1)*w+1+ystep;
			m += next_row_ofs;
			uint tmp = b[1];
			b[1] = b[0]+(xstep-1)*w+1+xstep;
			b[0] = tmp-w+1;

			c1 += next_row_ofs; c2 += next_row_ofs;	c3 += next_row_ofs; c4 += next_row_ofs;
		}

		xy_dist_sq += xy_dist_sq;	/* next level has 2x distance*/
		odd_start += odd_start;
	}
}


/* convert (row major) pp_v to (embedded quadtree) v */
static void traverseR(uint i, uint j, uint ml, uint ma, uint mr, uint l)
{
start:
	const uint mm = (ml+mr) >> 1;

	/* fill v[m] */
	v[j].dz = (signed char)((int)h[mm] - (((int)h[ml]+(int)h[mr])/2));
	v[j].e = pp_v[mm].e;
	v[j].r = pp_v[mm].r;

	/* recurse for left/right children */
	if(--l)		/* last level? */
	{			/* no - recurse */
		int a = (i+i)+j, b = (i << 2)-11;
		traverseR(j, ((a-10)&3)+b, ml, mm, ma, l);
		/* TRE: traverseR(j, ((a-9)&3)+b, ma, mm, mr, l); */
		i = j;
		j = ((a-9)&3)+b;
		ml = ma;
		ma = mm;
		goto start;
	}
}


int lod_preproc(const uint width, uchar* hmap)
{
	int i;

	w = width;
	h = hmap;

	/* make sure width == 2^(depth/2) + 1, depth >= 4 */
	if(((w-1) & (w-2)) || w < 5)
		return -1;

	depth = 2*log2(w-1);
	v_entries = (int)(w*w*4./3.);	/* ~33% wasted ghost nodes */

	int x[] = {0, w-1, w-1, 0, w/2, 0, w/2, w-1, w/2};
	int y[] = {0, 0, w-1, w-1, w/2, w/2, 0, w/2, w-1};
	for(i = 0; i < 9; i++)
	{
		vp[i].x = (short)x[i]; vp[i].y = (short)y[i];
		vp[i].z = h[x[i]+y[i]*w];
	}

	res_free(v);

	pp_v = (PP_V*)calloc(v_entries, sizeof(PP_V));
	v = (V*)res_malloc(v_entries * sizeof(V));
	if(!pp_v || !v)
		return -1;

	calc_e_r();

	/*
	 * convert (row major) pp_v to (interleaved quadtree) v
	 *
	 * start triangles:
	 * +-+-+
	 * |\|/|
	 * +-+-+
	 * |/|\|
	 * +-+-+
	 */

	uint m[9];	/* row major layout indices for vp[i] */
	for(i = 0; i < 9; i++)
		m[i] = (uint)(vp[i].x+vp[i].y*w);

	for(uint n = 0; n < 4; n++)
	{
		uint j = (n+1)%4, i = j+5;
		traverseR(i, 21-3*n, m[4], m[i], m[n], depth-2);
		traverseR(i, 21-3*j, m[j], m[i], m[4], depth-2);
	}

	free(pp_v);

	init();

	return 0;
}


int lod_load_preproc_data(const char* filename)
{
	int size;
	uchar* file = (uchar*)res_load(filename, &size);
	if(!file)
		return -1;

	memcpy(vp, file, sizeof(vp));

	res_free(v);
	v = (V*)(file+sizeof(vp));

	w = (int)sqrt(size / sizeof(V) / 4.*3.);

	/* make sure width == 2^(depth/2) + 1, depth >= 4 */
	if(((w-1) & (w-2)) || w < 5)
		return -1;

	depth = 2*log2(w-1);

	return init();
}


int lod_save_preproc_data(const char* filename)
{
	if(!depth)
		return -1;

	FILE* f = fopen(filename, "wb");
	if(!f)
		return errno;

	if(fwrite(vp, sizeof(vp), 1, f) != 1)
		return -1;

	if(fwrite(v, sizeof(V), v_entries, f) != v_entries)
		return -1;

	fclose(f);

	return 0;
}



/*****************************************************************************
 * refinement
 *****************************************************************************/

/* (normalized) view frustum plane equations */
float frustum[6][4];

/* eye pos (calculated from OpenGL modelview/projection matrices) */
float E[3];

/* tstrip data */
uint parity;
uint last_idx;
uint l1, l2;

/* vbs (triple buffer) */
static float* _vb[3];
static uint* _idx[3];
static uint write_vb, read_vb, ready_vb;

/* sync info */
static int new_vb_ready;
static uint fence[3];
static enum
{
	AVAIL,
	IN_USE,	/* writing to / reading from this vb */
	READY,	/* write complete, we'll read from this vb next frame */
	WAIT	/* waiting for DMA transfer to complete after last read */
}
vb_state[3];

/* pointers to current write vb */
uint* idx;
float* vb;


/* extract eye pos + view frustum from OpenGL modelview/projection matrices */
void lod_latch_view_pos()
{
	float proj[16], mv[16];		/* (column major!) matrices from OpenGL*/
	float m[4][4];				/* (row major!) proj*mv */
	float t;
	float l;
	int i, j, k;

	glGetFloatv(GL_MODELVIEW_MATRIX, mv);
	glGetFloatv(GL_PROJECTION_MATRIX, proj);

	/* view pos */
	memset(E, 0, sizeof(E));
	for(i = 0; i < 3; i++) 
		for(j = 0; j < 3; j++)
			E[i] -= (mv[4*i + j] * mv[12 + j]);

	/* m = proj*mv */
	memset(m, 0, sizeof(m));
	for(i = 0; i < 4; i++)
		for(j = 0; j < 4; j++)
			for(k = 0; k < 4; k++)
				m[i][j] += proj[4*k + i] * mv[4*j + k];

	/* frustum plane equations */
	for(i = 0; i < 6; i++)
	{
		k = i >> 1;
		l = 0.f;
		for(j = 0; j < 4; j++)
		{
			t = ((i & 1)? +m[k][j] : -m[k][j]) - m[3][j];
			frustum[i][j] = t;
			l += t*t;
		}

		/* normalize */
		for(j = 0; j < 4; j++)
			frustum[i][j] /= l;
	}
}


/*****************************************************************************
 * refine_c
 *****************************************************************************/

static inline void tstrip_append(const uint i, const uint p, const VP& v)
{
	if(i == l1 || i == l2)
		return;

	if(parity == p)
	{
		*idx = idx[-2];
		idx++;
	}
	else
	{
		parity = p;
		l2 = l1;
	}
	l1 = i;

	*idx++ = last_idx++;
	vb[0] = (float)v.x; vb[1] = (float)v.y;
	vb[2] = v.z;
	vb += 4;	/* V3F, stride = 16 */
}


static void refineR(uint i, uint j, VP vl, VP va, VP vr, uint32 l_vis)
{
	{	/* avoid VC++ "local var. init. skipped" error by
		   making locals go out of scope after "goto skip" */
start:

	if(((--l_vis)&0xffff) == 0)		// if(--l == 0)
		goto skip;

	const float e = (float)v[j].e, r = (float)v[j].r, z = (float)v[j].dz;

	VP vm; vm.x = (vl.x+vr.x)/2; vm.y = (vl.y+vr.y)/2; vm.z = (vl.z+vr.z)/2;

	/* MORPH / ACTIVE */
	float dx = vm.x-E[0], dy = vm.y-E[1], dz = vm.z-E[2];
	float d = dx*dx + dy*dy + dz*dz;
	float dmax = e*nu[0] + r; dmax *= dmax;

	dmax -= d;
	if(*(int32*)&dmax <= 0)	/* IEEE-754 specific: if(dmax <= 0) ... */
		goto skip;			/*   faster than fcomp, fnstsw */
	float dmin = e*nu[1] + r; dmin *= dmin;
	d -= dmin;
	vm.z += (*(uint32*)&d <= 0x80000000)? z * dmax / (dmax+d) : z; /* (d >= 0)? ... : z */

	/* CULL */
	if(l_vis < 0x001f0000)		/* not completely inside all planes; NOTE: far plane is ignored */
		for(uint n = 0, bit = (1 << 16); n < 5; n++, bit += bit)
			if(!(l_vis & bit))
			{
				d = vm.x*frustum[n][0] + vm.y*frustum[n][1] + vm.z*frustum[n][2] + frustum[n][3];
				if(d > r)		/* completely outside */
					goto skip;
				if(-d > r)		/* completely inside */
					l_vis |= bit;
			}

	/* VISIT CHILDREN (IN ORDER) */
	uint a = (i+i)+j, b = (i << 2)-11;
	refineR(j, ((a-10)&3)+b, va, vm, vr, l_vis);
	tstrip_append(i, l_vis&1, va);	/* index, parity, vertex coords */
	/* TRE: refineR(j, ((a-9)&3)+b, vl, vm, va, l, visible); */
	i = j;
	j = ((a-9)&3)+b;
	vr = va;
	va = vm;
	goto start;
	}

skip:
	tstrip_append(i, l_vis&1, va);	/* index, parity, vertex coords */
}


static void refine_c()
{
	int n, i, j;

	for(n = 0; n < 4; n++)
	{
		i = (n+1)%4+5, j = i-5;
		tstrip_append(n, 1, vp[n]);
		refineR(i, 21-3*n, vp[4], vp[i], vp[n], depth-2);
		tstrip_append(4, 0, vp[4]);
		refineR(i, 21-3*j, vp[j], vp[i], vp[4], depth-2);
	}
}


extern void refine_3dnow();

static SDL_Thread* thread;
static SDL_sem* sem;




/* refine speed measurement code */
#ifdef WIN32_PERF_MON

#include <windows.h>

static uint refines, swaps, frames;
static LARGE_INTEGER tris, ref_tris;
static LARGE_INTEGER start, end;

static void dump_perf_data()
{
	LARGE_INTEGER i;

	QueryPerformanceFrequency(&i);
	float freq = (float)i.QuadPart;
	
	QueryPerformanceCounter(&end);
	float tot_t = (end.QuadPart-start.QuadPart) / freq;

	FILETIME ft;
	GetThreadTimes(thread, &ft, &ft, &ft, (FILETIME*)&i);	/* only need user space time */
	float ref_t = i.QuadPart * 1e-7f;	/* convert FILETIME to seconds */

	int cpt = (int)(sys_cpu_freq / (ref_tris.QuadPart / ref_t));
	int tpf = (int)(tris.QuadPart / (float)frames);
	int fps = (int)(frames / tot_t);
	char buf[1000];
	sprintf(buf, "\nrefines: %d, swaps: %d, frames: %d\n"\
				 "clocks per tri: %d, refine throughput: %f Mtri/s\n"\
				 "tris/frame: %d, fps: %d, render throughput: %f Mtri/s\n", refines, swaps, frames, cpt, (ref_tris.QuadPart / ref_t)*1e-6f, tpf, fps, (tris.QuadPart / tot_t)*1e-6f);
	OutputDebugString(buf);
	printf(buf);
}

#endif




static int refine_thread_proc(void*)
{
	void(*refine)() = (sys_has_3dnow_pro)? refine_3dnow : refine_c;

#ifdef WIN32_PERF_MON
QueryPerformanceCounter(&start);
#endif

	for(;;)
	{
		/* find first unused vb */
		for(int i = 0; i < 3; i++)
			if(vb_state[i] == AVAIL)
			{
				write_vb = i;
				vb_state[i] = IN_USE;
				break;
			}

		/* init vb */
		vb = _vb[write_vb];
		idx = _idx[write_vb]+1;	/* skip dummy index */
		last_idx = 0;
		parity = 0;
		l1 = l2 = 0xffffffff;

		refine();

#ifdef WIN32_PERF_MON
refines++;
ref_tris.QuadPart += idx-_idx[write_vb];
#endif

		*idx = 0;

		vb_state[write_vb] = READY;
		ready_vb = write_vb;
		new_vb_ready = 1;

		SDL_SemWait(sem);
	}

	return 0;
}


/*****************************************************************************
 * init
 *****************************************************************************/

static void cleanup(void)
{
#ifdef WIN32_PERF_MON
dump_perf_data();
#endif

	SDL_KillThread(thread);
	SDL_DestroySemaphore(sem);

	if(use_VAR)
	{
		wglFreeMemoryNV(_vb[0]);
		glDeleteFencesNV(3, fence);
	}
}


static int init()
{
	/* already initialized? (init() is called by lod_preproc() and lod_load_preproc_data() */
	if(_vb[0])
		return 0;

	atexit(cleanup);

	use_VAR = ext_avail("GL_NV_vertex_array_range") &&
			  ext_avail("GL_NV_fence");

	/* alloc V3F vb (stride = 16) */
	const int vb_entries = 200*1024, idx_entries = 400*1024;
	if(use_VAR)
		_vb[0] = (float*)wglAllocateMemoryNV(vb_entries*sizeof(float)*4*3, .2f, .2f, .5f);
	if(!_vb[0])	/* VAR not supported or wglAllocateMemoryNV() failed */
	{
		use_VAR = 0;
		_vb[0] = (float*)res_malloc(vb_entries*sizeof(float)*4*3);
	}
	_vb[1] = _vb[0] + vb_entries*4;	/* 16 byte stride! */
	_vb[2] = _vb[1] + vb_entries*4;
	_idx[0] = (uint*)res_malloc(idx_entries*sizeof(uint)*1*3);
	_idx[1] = _idx[0] + idx_entries*1;
	_idx[2] = _idx[1] + idx_entries*1;
	if(!_vb[0] || !_idx[0])
		return -1;

	glEnableClientState(GL_VERTEX_ARRAY);

	if(use_VAR)
	{
		glGenFencesNV(3, fence);
		glVertexArrayRangeNV(vb_entries*sizeof(float)*4*3, _vb[0]);
		glEnableClientState(GL_VERTEX_ARRAY_RANGE_NV);
	}

	sem = SDL_CreateSemaphore(2);
	thread = SDL_CreateThread(refine_thread_proc, NULL);

	return 0;
}


void lod_render()
{
	static uint* r_idx;
	static uint r_cnt;

	if(--new_vb_ready == 0)	/* check if == 1 and reset */
	{
		vb_state[read_vb] = WAIT;
		if(use_VAR)
			glSetFenceNV(fence[read_vb], GL_ALL_COMPLETED_NV);
		
		read_vb = ready_vb;
		vb_state[read_vb] = IN_USE;
		
		r_idx = _idx[read_vb]+1; /* 1st index is skipped */
		r_cnt = idx-_idx[read_vb];
		glVertexPointer(3, GL_FLOAT, 16, (void*)_vb[read_vb]);

#ifdef WIN32_PERF_MON
swaps++;
#endif
	}

	/* mark no longer needed vbs as available */
	for(int i = 0; i < 3; i++)
		if(vb_state[i] == READY || (vb_state[i] == WAIT && (!use_VAR || glTestFenceNV(fence[i]))))
		{
			vb_state[i] = AVAIL;
			SDL_SemPost(sem);
		}

	glDrawElements(GL_TRIANGLE_STRIP, r_cnt, GL_UNSIGNED_INT, r_idx);

#ifdef WIN32_PERF_MON
frames++;
tris.QuadPart += r_cnt;
#endif
}


}		/* extern "C" { */