;____________________________________________________________________________________________________________________________
;
; Athlon XP-specific, MMX / 3DNow! Professional optimized implementation
; of "Visualization of Large Terrains Made Easy" (Lindstrom and Pascucci)
; with geomorphing, embedded quadtree indexing, view frustum culling,
; and multithreaded refinement
;
; Copyright (c) 2002 Jan Wassenberg
;
; This program is free software; you can redistribute it and/or
; modify it under the terms of the GNU General Public License as
; published by the Free Software Foundation; either version 2 of the
; License, or (at your option) any later version.
;
; This program is distributed in the hope that it will be useful, but
; WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
; General Public License for more details.
;
; Contact info:
;   Jan.Wassenberg@stud.uni-karlsruhe.de
;   http://www.stud.uni-karlsruhe.de/~urkt/
;____________________________________________________________________________________________________________________________
;


global _refine_3dnow

extern _nu
extern _E, _frustum
extern _depth
extern _v, _vp
extern _l1, _parity, _last_idx
extern _vb, _idx


[section .data use32]

align 64
_05		dd	0, 0.5			; used to average vl, vr
tmp		dd	0, 0
_1		dd	1.0

params		db	6, 7, 8, 5		; i param of refineR (params[n] == (n+1)%4+5)


[section .text use32]

;____________________________________________________________________________________________________________________________
;
; refineR - refine, morph and cull mesh; create output tstrip
;
; expected run time: 76 clocks (assuming memory access latency of 2 clocks)
;
; pseudocode:
;
;   refineR(i, j, vl, va, vr, l)
;     if(--l == 0)	// max refine depth reached?
;       goto skip	// yes - don't recurse
;
;     vm := midpoint(vl, vr)
;
;     // MORPH
;     d := distance(vm, camera)
;     dmin := v[j].e * nu_min + r, dmax := v[j].e * nu_max + r
;     if(d^2 >= dmax^2)	// vm is far enough away that it doesn't need to be split
;       goto skip
;     dz := v[j].dz
;     if(d^2 >= dmin^2)	// vm needs to be morphed from vm.z to vm.z+v[j].z
;       dz *= (dmax^2-d^2) / (dmax^2-dmin^2)
;     vm.z += dz
;
;     // CULL
;     foreach frustum plane i
;       if(distance(vm, plane i) > v[j].r)	// are vm and all children completely outside plane i?
;         goto skip				// yes - don't recurse
;
;     refineR(j, cl(i,j), va, vm, vr, l)
;     tstrip_append(i, l&1, va)			// append va (= v[i]) to tstrip with parity (l&1)
;     refineR(j, cr(i,j), vl, vm, va, l)
;
;     return
;
;     skip:
;       tstrip_append(i, l&1, va)		// append va (= v[i]) to tstrip with parity (l&1)
;
; notes:
;  - only the 2 most "cull effective" frustum planes (left, right) are checked
;    => no inside_plane bits necessary
;  - the culling code assumes the camera doesn't roll! (frustum[0..1].c == 0)
;
; regs:
; in:
;   ebx -> params
;   esi = l2  [static]
;   edi = vb  [static]
;   ebp = idx [static]
;   mm0 = vl.z | vl.y vl.x
;   mm2 = vr.z | vr.y vr.x
;   mm7 = ? | ? | e dz | r  (=v[j])
; destroyed:
;   e?x, mm?
;
; stack frame:
;   esp offset	length
;       0         4         return address
;       4         8         local: vm
;       8         36        params (see below)
;____________________________________________________________________________________________________________________________
;

%define vm ebx-8
%define i  ebx+0
%define j  ebx+4
%define vl ebx+8
%define va ebx+16
%define vr ebx+24
%define l  ebx+32

align 64
refineR:
	movq		mm1, mm0		; vl.z | ?
	paddw		mm0, mm2		; ? | vl.y+vr.y vl.x+vr.x
	pi2fw		mm6, mm7		; ? | r
	psllq		mm7, 8			; ? | 0 e | dz ? | ?
	dec		dword [l]		; --l == 0?
	jz		near skip		; yes - stop recursion
	pfadd		mm1, mm2		; vl.z+vr.z | ?
	psrlw		mm0, 1			; ? | vm.y vm.x
	movq		mm2, [_05]		; 0.5 | 0
	pi2fw		mm5, mm7		; e | ?
	mov		eax, [i]
	mov		edx, [j]
	movd		[vm], mm0		; save vm.x, vm.y as shorts
	punpcklwd	mm0, mm0		; ? vm.y | ? vm.x
	punpckldq	mm6, mm6		; r | r
	movq		mm4, [_nu]		; numin | numax
	pi2fw		mm0, mm0		; vm.y | vm.x
	pfmul		mm1, mm2		; vm.z | ?
	movq		mm2, [_E]		; E.y | E.x
	punpckhdq	mm5, mm5		; e | e
	lea		ecx, [edx+eax*2-10]
	movq		mm3, [_frustum]		; f0.ny | f0.nx
	pfmul		mm5, mm4		; emin | emax
	and		ecx, 3
	movq		mm4, [_frustum+16]	; f0.ny | f0.nx
	pfsub		mm2, mm0		; dy | dx
	pswapd		mm1, mm1		; ? | vm.z
	lea		ecx, [ecx+eax*4-11]	; cl(i,j)
	mov		eax, [l]
	movd		[vm+4], mm1		; save vm.z
	pfsub		mm1, [_E+8]		; ? | dz
	pfmul		mm3, mm0		; vm.y*f0.ny | vm.x*f0.nx
	pfadd		mm5, mm6		; dmin | dmax
	pfmul		mm4, mm0		; vm.y*f1.ny | vm.x*f1.nx
	pfmul		mm2, mm2		; dy^2 | dx^2
	pfadd		mm3, [_frustum+8]	; lo+hi == d(vm,f0)
	pfadd		mm4, [_frustum+24]	; lo+hi == d(vm,f1)
	pfmul		mm5, mm5		; dmin^2 | dmax^2
	pfacc		mm2, mm2		; ? | dy^2+dx^2
	pfmul		mm1, mm1		; ? | dz^2
	movq		mm0, mm5		; ? | dmax^2
	pfnacc		mm5, mm5		; ? | dmax^2-dmin^2
	pfadd		mm2, mm1		; ? | d^2
	pfacc		mm3, mm4		; d(vm,f1) | d(vm,f0)
	pfrcp		mm5, mm5		; ? | 1/(dmax^2-dmin^2)
	pfsubr		mm2, mm0		; ? | dmax^2-d^2
	movq		mm0, [va]
	pcmpgtd		mm3, mm6		; outside f1 | outside f0
	movq		mm6, mm7		; ? | ? | dz ? | ?
	packssdw	mm3, mm3		; ? | outside_f1 outside_f0
	movd		[tmp], mm2
	pfmul		mm5, mm2		; ? | (dmax^2-d^2)/(dmax^2-dmin^2)
	movd		mm4, [_1]		; ? | 1.f
	cmp		dword [tmp], 0		; dmax <= d?
	psrad		mm6, 24			; ? | dz (sign extended to 32 bits)
	movd		[tmp], mm3		; outside_f1 outside_f0
	movq		mm2, [vr]
	jle		near skip		; yes - stop recursion
	pi2fd		mm6, mm6		; ? | dz
	cmp		dword [tmp], 0		; vm outside of frustum plane 0 or 1?
	pfmin		mm5, mm4		; ? | z morph factor
	jnz		near skip		; yes - stop recursion
	sub		esp, 44			; esp -> new stack frame + 4
	lea		ebx, [esp+8]		; ebx -> new frame's params
	mov		[i], edx
	mov		[j], ecx
	pfmul		mm6, mm5		; ? | z offset
	movd		mm3, [48+vm+4]		; old vm.z
	movq		[vl], mm0
	movq		[vr], mm2
	mov		[l], eax
	mov		eax, [_v]
	pfadd		mm3, mm6		; vm.z = morphed z
	movd		mm1, [48+vm]		; old vm.y | old vm.y
	movd		mm7, [eax+ecx*4]	; ? | 0 | e dz | r    (= v[j])
	movd		[48+vm+4], mm3		; save vm.z
	punpckldq	mm1, mm3		; vm.z | vm.y vm.x
	movq		[va], mm1
	call		refineR
	lea		ebx, [esp+12]		; restore frame pointer
	mov		ecx, [ebp-8]		; idx[-2]
	movq		mm0, [vl]
	movq		mm3, [vm]
	mov		eax, [i]
	movq		mm2, [va]		; va.z | va.y va.x
	mov		[ebp], ecx		; append idx[-2] to idx (undone if _parity != p below)
	mov		ecx, [_l1]
	mov		edx, [l]
	cmp		eax, ecx		; i == l1?
	pswapd		mm1, mm2		; va.y va.x | va.z
	je		.1			; yes - skip tstrip_append
	cmp		eax, esi		; i == l2?
	je		.1			; yes - skip tstrip_append
	bt		edx, 0			; CF = l&1
	rcr		dword [_parity], 1	; OF = (_parity != p); _parity = p
	cmovo		esi, ecx		; if(_parity != p) l2 = l1
	mov		[_l1], eax		; l1 = i
	mov		ecx, [_last_idx]
	movntq		[edi+8], mm1		; write va.z to vb
	setno		dl			; edx = (_parity == p)
	punpckhwd	mm1, mm1		; ? | va.y | ? | va.x
	lea		ebp, [ebp+4+edx*4]	; idx += ((_parity == p)? 2 : 1)
	inc		dword [_last_idx]
	pi2fw		mm1, mm1		; va.y | va.x
	mov		[ebp-4], ecx		; append _last_idx++ to idx
	add		edi, 16			; vb++
	movntq		[edi-16], mm1		; write va.x, va.y to vb
align 16
.1:
	mov		edx, [j]
	movq		[va], mm3
	movq		[vr], mm2
	mov		[i], edx
	lea		ecx, [edx+eax*2-9]
	mov		edx, [_v]
	and		ecx, 3
	lea		ecx, [ecx+eax*4-11]	; cr(i,j)
	mov		[j], ecx
	movd		mm7, [edx+ecx*4]	; ? | 0 | e dz | r    (= v[j])
	jmp		refineR
align 64
skip:
	mov		eax, [i]
	mov		edx, [l]
	movq		mm0, [va]		; va.z | va.y va.x
tstrip_append:
	mov		ecx, [ebp-8]		; idx[-2]
	mov		[ebp], ecx		; append idx[-2] to idx (undone if _parity != p below)
	mov		ecx, [_l1]
	cmp		eax, ecx		; i == l1?
	pswapd		mm1, mm0		; va.y va.x | va.z
	je		.2			; yes - skip tstrip_append
	cmp		eax, esi		; i == l2?
	je		.2			; yes - skip tstrip_append
	bt		edx, 0			; CF = l&1
	rcr		dword [_parity], 1	; OF = (_parity != p); _parity = p
	cmovo		esi, ecx		; if(_parity != p) l2 = l1
	mov		[_l1], eax		; l1 = i
	mov		ecx, [_last_idx]
	movntq		[edi+8], mm1		; write va.z to vb
	setno		dl			; edx = (_parity == p)
	punpckhwd	mm1, mm1		; ? | va.y | ? | va.x
	lea		ebp, [ebp+4+edx*4]	; idx += ((_parity == p)? 2 : 1)
	inc		dword [_last_idx]
	pi2fw		mm1, mm1		; va.y | va.x
	mov		[ebp-4], ecx		; append _last_idx++ to idx
	add		edi, 16			; vb++
	movntq		[edi-16], mm1		; write va.x, va.y to vb

align 16
.2:
	ret		44


;____________________________________________________________________________________________________________________________
;
; _refine_3dnow
;____________________________________________________________________________________________________________________________
;
_refine_3dnow:
	pushad

	or		esi, -1
	mov		[_l1], esi		; l1 = l2 = -1
	xor		eax, eax
	mov		[_parity], eax
	mov		[_last_idx], eax
	mov		edi, [_vb]
	mov		ebp, [_idx]

	push		byte -3			; (local) n = -3

.1:	mov		ecx, [esp]

	; TSTRIP
	lea		eax, [ecx+3]		; n
	mov		edx, 1
	movq		mm0, [_vp+eax*8]	; v[n]
	sub		esp, 44
	call		tstrip_append

	mov		ecx, [esp]		; n

	; REFINER
	movzx		eax, byte [params+ecx+3]; i
	lea		edx, [ecx*2+ecx-12]	; -(21-3*n)
	neg		edx
	movq		mm0, [_vp+4*8]
	movq		mm1, [_vp+eax*8]
	movq		mm2, [_vp+ecx*8+24]
	mov		ecx, [_depth]
	sub		ecx, 2
	sub		esp, 44
	lea		ebx, [esp+8]
	mov		[i], eax
	mov		[j], edx
	movq		[vl], mm0
	movq		[va], mm1
	movq		[vr], mm2
	mov		[l], ecx

	mov		eax, [_v]
	movd		mm7, [eax+edx*4]	; ? | 0 | e dz | r    (= v[j])

	call		refineR

	; TSTRIP
	mov		eax, 4
	mov		edx, 0
	movq		mm0, [_vp+4*8]
	sub		esp, 44
	call		tstrip_append

	mov		ecx, [esp]

	; REFINER
	movzx		eax, byte [params+ecx+3]; i
	lea		edx, [eax*2+eax-36]	; -(21-3*(i-5))
	neg		edx
	mov		ecx, [_depth]
	sub		ecx, 2
	movq		mm0, [_vp+eax*8-40]
	movq		mm1, [_vp+eax*8]
	movq		mm2, [_vp+4*8]
	sub		esp, 44
	lea		ebx, [esp+8]
	mov		[i], eax
	mov		[j], edx
	movq		[vl], mm0
	movq		[va], mm1
	movq		[vr], mm2
	mov		[l], ecx

	mov		eax, [_v]
	movd		mm7, [eax+edx*4]	; ? | 0 | e dz | r    (= v[j])

	call		refineR

	inc		dword [esp]		; n++
	jle		.1			; if(n <= 0) goto .1

	pop		edx			; free local variable n

	mov		[_vb], edi
	mov		[_idx], ebp

	popad

	femms

	ret