/*
 * emulation of a subset of the SDL API for Win32
 *
 * Copyright (c) 2002 Jan Wassenberg
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Contact info:
 *   Jan.Wassenberg@stud.uni-karlsruhe.de
 *   http://www.stud.uni-karlsruhe.de/~urkt/
 */

#include "precompiled.h"

#include "wsdl.h"


/************************************************************************************************
 * events / input
 ************************************************************************************************/

int SDL_PollEvent(SDL_Event* ev)	/* ev must be valid */
{
	/* windows messages */
	MSG msg;
	while(PeekMessage(&msg, 0, 0, 0, PM_REMOVE))
	{
		/* keyboard */
		if(msg.message == WM_KEYUP || msg.message == WM_KEYDOWN)
		{
			ev->type = (msg.message == WM_KEYDOWN)? SDL_KEYDOWN : SDL_KEYUP;
			ev->key.keysym.sym = (SDLKey)msg.wParam;
			return 1;
		}

		DispatchMessage(&msg);
	}

	/* mouse */
	static int x, y;
	POINT p;
	GetCursorPos(&p);
	if(x != p.x || y != p.y)
	{
		ev->type = SDL_MOUSEMOTION;
		ev->motion.x = x = p.x;
		ev->motion.y = y = p.y;
		return 1;
	}

	return 0;
}


/************************************************************************************************
 * threads
 ************************************************************************************************/

/*
 * can't run threads directly because win32 starts them with __stdcall.
 *
 * FIXME: parameter passing not implemented
 */
static DWORD WINAPI run_thread(void* param)
{
	return (DWORD)((SDL_THREAD_PROC)param)(NULL);
}


/* param is ignored! */
SDL_Thread* SDL_CreateThread(SDL_THREAD_PROC proc, void* /* param */)
{
	return (SDL_Thread*)CreateThread(0, 0, run_thread, proc, 0, 0);
}


/************************************************************************************************
 * video
 ************************************************************************************************/

static HWND hWnd;
static HDC hDC;
static HGLRC hGLRC;
static DEVMODE dm;
static int fullscreen;


/*
 * set video mode wxh:bpp if necessary.
 *
 * shortcuts:
 *   DEVMODE.dmSize is set to 0
 *   PIXELFORMATDESCRIPTOR.nSize and nVersion are set to 0
 */
int SDL_SetVideoMode(int w, int h, int bpp, uint32 flags)
{
	fullscreen = (flags & SDL_FULLSCREEN);

	EnumDisplaySettings(NULL, ENUM_CURRENT_SETTINGS, &dm);
	int _w = dm.dmPelsWidth, _h = dm.dmPelsHeight, _bpp = dm.dmBitsPerPel;
	if(bpp != _bpp || w > _w || h > _h ||
	  (fullscreen && (w != _w || h != _h)))
	{
		dm.dmPelsWidth  = w;
		dm.dmPelsHeight = h;
		dm.dmBitsPerPel = bpp;
		dm.dmFields = DM_PELSWIDTH|DM_PELSHEIGHT|DM_BITSPERPEL;
	}
	ChangeDisplaySettings(&dm, fullscreen? CDS_FULLSCREEN : 0);

	SetWindowPos(hWnd, NULL, 0, 0, w, h, SWP_NOCOPYBITS|SWP_NOZORDER|SWP_NOREDRAW);

	hDC = GetDC(hWnd);

	/* set pixel format */
	static PIXELFORMATDESCRIPTOR pfd;
	pfd.dwFlags = PFD_SUPPORT_OPENGL|PFD_DRAW_TO_WINDOW|PFD_DOUBLEBUFFER;
	pfd.cColorBits = (BYTE)bpp;
	if(!SetPixelFormat(hDC, ChoosePixelFormat(hDC, &pfd), &pfd))
		return 0;

	hGLRC = wglCreateContext(hDC);
	wglMakeCurrent(hDC, hGLRC);

	return 1;
}


void SDL_GL_SwapBuffers()
{
	SwapBuffers(hDC);
}


/************************************************************************************************
 * init
 ************************************************************************************************/

static LRESULT CALLBACK wndproc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch(uMsg)
	{
	case WM_ACTIVATEAPP:
		if(wParam)
		{
			ChangeDisplaySettings(&dm, fullscreen? CDS_FULLSCREEN : 0);
			ShowWindow(hWnd, SW_RESTORE);
			wglMakeCurrent(hDC, hGLRC);
		}
		else
		{
			wglMakeCurrent(0, 0);
			ShowWindow(hWnd, SW_MINIMIZE);
			ChangeDisplaySettings(0, 0);
		}
		return 0;

	default:
		return DefWindowProc(hWnd, uMsg, wParam, lParam);
	}
}


static float perf_cnt_to_ms;

int SDL_Init(uint32 flags)
{
	/* redirect stdout */
	freopen("stdout.txt", "w", stdout);

	/* init timer */
	LARGE_INTEGER freq;
	QueryPerformanceFrequency(&freq);
	perf_cnt_to_ms = 1000.f / freq.QuadPart;

	/* register window class */
	static WNDCLASS wc;
	wc.style = CS_OWNDC;
	wc.lpfnWndProc  = wndproc;
	wc.lpszClassName = "SDL";
	RegisterClass(&wc);

	hWnd = CreateWindowEx(WS_EX_TOPMOST*0, "SDL", "SDL", WS_POPUP, 0, 0, 1024, 768, 0, 0, 0, 0);

	return 0;
}


void SDL_Quit()
{
	fclose(stdout);

	ChangeDisplaySettings(0, 0);
}






uint32 SDL_GetTicks()
{
	LARGE_INTEGER i;
	QueryPerformanceCounter(&i);
	return (int)(i.QuadPart * perf_cnt_to_ms);
}

